import os
from py_book_util import util


class GitRemoteBook(object):
    def __init__(self, remote_file, git_repo_path, local_dir, current_working_dir):
        self.remote_file = remote_file
        self.git_repo_path = git_repo_path
        self.local_dir = local_dir
        self.current_working_dir = current_working_dir

    def get_remote_path(self):
        return self.git_repo_path + "/" + self.remote_file

    def get_git_repo_name(self):
        return self.git_repo_path[self.git_repo_path.rindex("/") + 1 : :]

    def get_local_path(self):
        remote_path = self.get_remote_path()
        local_path = self.local_dir + "/" + remote_path
        if remote_path.startswith("/"):
            local_path = self.local_dir + remote_path
        print("local_path is %s" % local_path)
        return local_path

    def get_local_path_in_book(self):
        remote_path = self.get_remote_path()
        local_path_in_book = remote_path
        if remote_path.startswith("/"):
            local_path_in_book = remote_path[1::]
        print("local_path_in_book is %s" % local_path_in_book)
        return local_path_in_book

    def download_file_from_remote_repo_to_local_dir(self):
        import shutil

        if os.path.isdir(self.local_dir):
            import stat

            def delete_on_error(func, path, execinfo):
                os.chmod(path, stat.S_IWUSR)
                func(path)

            shutil.rmtree(self.local_dir, onerror=delete_on_error)
        os.mkdir(self.local_dir)
        cucumber_task = "Download %s %s from server" % (
            self.get_git_repo_name(),
            self.remote_file,
        )
        util.print_run_cmd('006_cucumber_parallel "%s"' % cucumber_task)
        print("\n")
        if os.path.isfile(self.get_local_path()):
            util.run_deno_cmd("dos2unix " + self.get_local_path())
            util.run_deno_cmd("md5sum " + self.get_local_path())
            return True
        else:
            print("Cannot find %s. Please download again!\n" % self.get_local_path())
            return False

    def upload_file_from_local_dir_to_remote_repo(self):
        cucumber_task = "Upload %s %s to server" % (
            self.get_git_repo_name(),
            self.remote_file,
        )
        util.run_deno_cmd("dos2unix " + self.get_local_path())
        util.print_run_cmd('006_cucumber_parallel "%s"' % cucumber_task)

    def book_git_init(self):
        os.chdir(self.current_working_dir)
        os.chdir(self.local_dir)
        assert os.getcwd().endswith("book")
        util.print_run_cmd("git init")
        util.print_run_cmd("git config core.autocrlf input")
        os.chdir(self.current_working_dir)

    def book_git_add(self):
        os.chdir(self.current_working_dir)
        os.chdir(self.local_dir)
        assert os.getcwd().endswith("book")
        util.print_run_cmd("git add " + self.get_local_path_in_book())
        os.chdir(self.current_working_dir)

    def book_git_status(self):
        os.chdir(self.current_working_dir)
        os.chdir(self.local_dir)
        assert os.getcwd().endswith("book")
        util.print_run_cmd('git status"')
        os.chdir(self.current_working_dir)

    def book_git_commit(self, message):
        os.chdir(self.current_working_dir)
        os.chdir(self.local_dir)
        assert os.getcwd().endswith("book")
        util.print_run_cmd('git commit -m "%s"' % message)
        os.chdir(self.current_working_dir)

    def book_checkout_and_check_md5sum(self, md5sum):
        file_name_in_book = self._book_prepare()
        assert md5sum == util.get_file_md5sum(file_name_in_book)
        print("file_name_downloaded is %s" % file_name_in_book)
        # util.display_code(file_name_in_book)
        # from IPython.display import Code
        # Code(filename=file_name_in_book, language="text")
        os.chdir(self.current_working_dir)

    def book_move_line(self, old_line, new_line, ori_md5, new_md5):
        file_name_in_book = self._book_prepare()
        if ori_md5 == util.get_file_md5sum(file_name_in_book):
            util.move_line_in_file(file_name_in_book, old_line, new_line)
            commit_message = "Move line %d to line %d" % (old_line, new_line)
            self._book_commit(file_name_in_book, commit_message, new_md5)
        os.chdir(self.current_working_dir)

    def book_move_line_block(self, line_begin, line_end, new_line, ori_md5, new_md5):
        file_name_in_book = self._book_prepare()
        if ori_md5 == util.get_file_md5sum(file_name_in_book):
            util.move_line_block_in_file(
                file_name_in_book, line_begin, line_end, new_line
            )
            commit_message = "Move line block(%d, %d) to line %d" % (
                line_begin,
                line_end,
                new_line,
            )
            self._book_commit(file_name_in_book, commit_message, new_md5)
        os.chdir(self.current_working_dir)

    def book_del_lines(self, line_numbers, ori_md5, new_md5):
        if isinstance(line_numbers, list) == False:
            line_numbers = [line_numbers]
        line_numbers.sort()
        line_numbers.reverse()
        file_name_in_book = self._book_prepare()
        if ori_md5 == util.get_file_md5sum(file_name_in_book):
            for single_line_num in line_numbers:
                util.del_line_in_file(file_name_in_book, single_line_num)
            commit_message = "Del line %s" % str(line_numbers)
            self._book_commit(file_name_in_book, commit_message, new_md5)
        os.chdir(self.current_working_dir)

    def book_del_blocks(self, line_blocks, ori_md5, new_md5):
        if isinstance(line_blocks, dict) == False:
            line_blocks = {line_blocks: line_blocks}
        # {1:3, 4:4}
        # 4 --> {4:4}
        # {1:3}
        tmp_line_numbers = line_blocks.keys()
        line_numbers = []
        for single_num in tmp_line_numbers:
            line_numbers.append(single_num)
        line_numbers.sort()
        line_numbers.reverse()
        file_name_in_book = self._book_prepare()
        if ori_md5 == util.get_file_md5sum(file_name_in_book):
            for single_line_num in line_numbers:
                util.del_block_in_file(
                    file_name_in_book, single_line_num, line_blocks[single_line_num]
                )
            commit_message = "Del block %s" % str(line_blocks)
            self._book_commit(file_name_in_book, commit_message, new_md5)
        os.chdir(self.current_working_dir)

    def book_remove_indent(
        self, block_line_begin, block_line_end, ori_md5, new_md5, indent_space_number=4
    ):
        file_name_in_book = self._book_prepare()
        if ori_md5 == util.get_file_md5sum(file_name_in_book):
            util.remove_indent_space_in_file(
                file_name_in_book, block_line_begin, block_line_end, indent_space_number
            )
            commit_message = "Remove indent (%d space char) from line %d to line %d" % (
                indent_space_number,
                block_line_begin,
                block_line_end,
            )
            self._book_commit(file_name_in_book, commit_message, new_md5)
        os.chdir(self.current_working_dir)

    def book_rstrip_lines(self, line_numbers, ori_md5, new_md5):
        if isinstance(line_numbers, list) == False:
            line_numbers = [line_numbers]
        file_name_in_book = self._book_prepare()
        print(util.get_file_md5sum(file_name_in_book))
        if ori_md5 == util.get_file_md5sum(file_name_in_book):
            for single_line_num in line_numbers:
                util.rstrip_space_in_file(file_name_in_book, single_line_num)
            commit_message = "Remove white space at end of line %s" % str(line_numbers)
            self._book_commit(file_name_in_book, commit_message, new_md5)
        os.chdir(self.current_working_dir)

    def book_append_content(self, file_content, commit_message, ori_md5, new_md5):
        file_name_in_book = self._book_prepare()
        if ori_md5 == util.get_file_md5sum(file_name_in_book):
            util.append_text_file(file_name_in_book, file_content)
            self._book_commit(file_name_in_book, commit_message, new_md5)
        os.chdir(self.current_working_dir)

    def _book_prepare(self):
        os.chdir(self.current_working_dir)
        os.chdir(self.local_dir)
        assert os.getcwd().endswith("book")
        file_name_in_book = self.get_local_path_in_book()
        os.remove(file_name_in_book)
        util.print_run_cmd("git checkout %s" % file_name_in_book)
        print(util.get_file_md5sum(file_name_in_book))
        return file_name_in_book

    def _book_commit(self, file_name_in_book, commit_message, new_md5):
        util.run_deno_cmd("dos2unix " + file_name_in_book)
        util.print_run_cmd("git diff " + file_name_in_book)
        print(util.get_file_md5sum(file_name_in_book))
        assert new_md5 == util.get_file_md5sum(file_name_in_book)
        util.print_run_cmd("git add " + file_name_in_book)
        util.print_run_cmd('git commit -m "%s"' % commit_message)

    def book_replace_lines(self, line_dict, commit_message, ori_md5, new_md5):
        line_numbers = line_dict.keys()
        file_name_in_book = self._book_prepare()
        if ori_md5 == util.get_file_md5sum(file_name_in_book):
            for single_line_num in line_numbers:
                util.replace_line_in_file(
                    file_name_in_book, single_line_num, line_dict[single_line_num]
                )
            self._book_commit(file_name_in_book, commit_message, new_md5)
        os.chdir(self.current_working_dir)

    def book_add_line(self, line_num, file_content, commit_message, ori_md5, new_md5):
        file_name_in_book = self._book_prepare()
        if ori_md5 == util.get_file_md5sum(file_name_in_book):
            util.add_line_in_file(file_name_in_book, line_num, file_content)
            self._book_commit(file_name_in_book, commit_message, new_md5)
        os.chdir(self.current_working_dir)

    def book_replace_line(
        self, line_num, file_content, commit_message, ori_md5, new_md5
    ):
        file_name_in_book = self._book_prepare()
        if ori_md5 == util.get_file_md5sum(file_name_in_book):
            util.replace_line_in_file(file_name_in_book, line_num, file_content)
            self._book_commit(file_name_in_book, commit_message, new_md5)
        os.chdir(self.current_working_dir)



# https://www.cnblogs.com/superbaby11/p/15560994.html
from threading import Thread


def run_cmd(cmd):
    """Run shell command and return command output.

    Examples:
        >>> run_cmd("cd ..")
        ''

    Args:
        cmd (string): A shell command.

    Returns:
        string: Command output.
    """
    import subprocess

    return subprocess.getoutput(cmd)


def run_deno_cmd(cmd):
    from pathlib import Path

    return run_cmd(str(Path.home()) + "/.deno/bin/" + cmd)


def print_run_cmd(cmd):
    import subprocess

    # http://blog.kagesenshi.org/2008/02/teeing-python-subprocesspopen-output.html
    p = subprocess.Popen(
        cmd,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        encoding="utf-8",
        errors="replace",
    )
    stdout_buffer = []
    while True:
        line = p.stdout.readline()
        stdout_buffer.append(line)
        print(line, end="")
        if line == "" and p.poll() != None:
            break
    return "".join(stdout_buffer)


def create_text_file(file_name, file_content):
    with open(file_name, "w") as file:
        file.write(file_content)


# https://www.jb51.net/article/232129.htm
class MyThread(Thread):
    def __init__(
        self, group=None, target=None, name=None, args=(), kwargs=None, *, daemon=None
    ):
        super().__init__()
        self.fun = target
        self.args = args
        self.kwargs = kwargs

    def run(self):
        self.fun(*self.args, **self.kwargs)


def thread_run_cmd(cmd):
    t = MyThread(target=run_cmd, args=(cmd,), kwargs={})
    t.start()
    # t.join()


def thread_run_deno_cmd(cmd):
    t = MyThread(target=run_deno_cmd, args=(cmd,), kwargs={})
    t.start()
    # t.join()


def get_page_content(url):
    # https://www.cnblogs.com/herbert/p/10789343.html
    # https://www.selenium.dev/documentation/webdriver/drivers/remote_webdriver/#tabs-1-1
    from selenium import webdriver
    import time
    import os

    # https://blog.csdn.net/ad72182009/article/details/116117744
    if "http_proxy" in os.environ:
        del os.environ["http_proxy"]
    chrome_options = webdriver.ChromeOptions()
    driver = webdriver.Remote(
        command_executor="http://127.0.0.1:4444/wd/hub", options=chrome_options
    )
    driver.get(url=url)
    # Wait some time for page loading
    time.sleep(4)
    found_page_source = driver.page_source
    driver.close()
    return found_page_source


def display_code(content):
    from IPython.display import Code

    return Code(data=content)


def display_iframe(url):
    from IPython.display import IFrame

    return IFrame(src=url, width="100%", height=300)


def get_file_md5sum(file):
    import hashlib
    import os

    if not os.path.exists(file):
        return "md5sum: {}: No such file or directory".format(file)
    elif os.path.isdir(file):
        return "md5sum: {}: Is a directory".format(file)
    elif os.path.isfile(file):
        with open(file, "rb") as fd:
            data = fd.read()
        # return "{}  {}".format(hashlib.md5(data).hexdigest(), file)
        return hashlib.md5(data).hexdigest()
    else:
        return "md5sum: {}: Unexpected error".format(file)


def get_str_md5sum(content):
    import hashlib

    return hashlib.md5(content.encode("utf8")).hexdigest()


def append_text_file(file_name, file_content):
    with open(file_name, "a") as file:
        file.write(file_content)


def replace_line_in_file(file_path, old_line_number, new_line_content):
    import fileinput

    line_number = 0
    with fileinput.FileInput(file_path, inplace=True) as f:
        for line in f:
            line_number = line_number + 1
            if line_number == old_line_number:
                print(new_line_content, end="")
            else:
                print(line, end="")


def insert_line_at_top_of_file(file_path, new_line_content, new_line_content_end="\n"):
    import fileinput

    flag_inserted = False
    with fileinput.FileInput(file_path, inplace=True) as f:
        for line in f:
            if flag_inserted == False:
                print(new_line_content, end=new_line_content_end)
                flag_inserted = True
            print(line, end="")


def add_line_in_file(
    file_path, old_line_number, new_line_content, new_line_content_end="\n"
):
    import fileinput

    line_number = 0
    with fileinput.FileInput(file_path, inplace=True) as f:
        for line in f:
            line_number = line_number + 1
            if line_number == old_line_number:
                print(new_line_content, end=new_line_content_end)
            print(line, end="")


def del_line_in_file(file_path, old_line_number):
    import fileinput

    deleted_line = ""
    line_number = 0
    with fileinput.FileInput(file_path, inplace=True) as f:
        for line in f:
            line_number = line_number + 1
            if line_number == old_line_number:
                deleted_line = line
            else:
                print(line, end="")
    return deleted_line


def del_line_block_in_file(file_path, old_line_number_begin, old_line_number_end):
    import fileinput

    deleted_line = ""
    line_number = 0
    with fileinput.FileInput(file_path, inplace=True) as f:
        for line in f:
            line_number = line_number + 1
            if old_line_number_begin == line_number:
                deleted_line = line
            else:
                if (
                    old_line_number_begin < line_number
                    and line_number <= old_line_number_end
                ):
                    deleted_line = deleted_line + line
                else:
                    print(line, end="")
    return deleted_line


def move_line_in_file(file_path, old_line_number, new_line_number):
    if old_line_number <= new_line_number and new_line_number <= (old_line_number + 1):
        return ""
    else:
        new_line_content = del_line_in_file(file_path, old_line_number)
        if old_line_number > new_line_number:
            add_line_in_file(file_path, new_line_number, new_line_content, "")
        else:
            add_line_in_file(file_path, new_line_number - 1, new_line_content, "")


def move_line_block_in_file(
    file_path, old_line_number_begin, old_line_number_end, new_line_number
):
    if old_line_number_begin <= new_line_number and new_line_number <= (
        old_line_number_end + 1
    ):
        return
    else:
        new_line_content = del_line_block_in_file(
            file_path, old_line_number_begin, old_line_number_end
        )
        if old_line_number_begin > new_line_number:
            add_line_in_file(file_path, new_line_number, new_line_content, "")
        else:
            add_line_in_file(
                file_path,
                new_line_number - (old_line_number_end - old_line_number_begin + 1),
                new_line_content,
                "",
            )


def read_text_file(file_path):
    from pathlib import Path

    return Path(file_path).read_text()


def display_text(content):
    from IPython.display import Code

    return Code(data=content, language="text")


def remove_indent_space_in_file(
    file_path, block_line_begin, block_line_end, indent_space_number
):
    import fileinput

    line_number = 0
    with fileinput.FileInput(file_path, inplace=True) as f:
        for line in f:
            line_number = line_number + 1
            if line_number >= block_line_begin and line_number <= block_line_end:
                if len(line) < indent_space_number:
                    print(line, end="")
                else:
                    new_line_content = line[indent_space_number:]
                    print(new_line_content, end="")
            else:
                print(line, end="")


def rstrip_space_in_file(file_path, line_number):
    import fileinput

    cur_line_number = 0
    with fileinput.FileInput(file_path, inplace=True) as f:
        for line in f:
            cur_line_number = cur_line_number + 1
            if line_number == cur_line_number:
                print(line.rstrip() + "\n", end="")
            else:
                print(line, end="")


def display_image(img_file):
    from IPython.display import Image

    return Image(img_file)


def del_block_in_file(file_path, block_line_begin, block_line_end):
    import fileinput

    line_number = 0
    with fileinput.FileInput(file_path, inplace=True) as f:
        for line in f:
            line_number = line_number + 1
            if line_number >= block_line_begin and line_number <= block_line_end:
                pass
            else:
                print(line, end="")


def full_page_snapshot(url, dir_snapshot):
    from selenium import webdriver
    from selenium.webdriver.common.by import By
    import time
    import os

    assert os.path.isdir(dir_snapshot)
    if "http_proxy" in os.environ:
        del os.environ["http_proxy"]
    chrome_options = webdriver.ChromeOptions()
    driver = webdriver.Remote(
        command_executor="http://127.0.0.1:4444/wd/hub", options=chrome_options
    )
    driver.maximize_window()
    driver.get(url=url)
    # Wait some time for page loading
    time.sleep(4)
    found_page_source = driver.page_source
    try:
        span_collapse = driver.find_element(by=By.XPATH, value='//span[text()="«"]')
        span_collapse.click()
    except Exception as e:
        print(e)
    driver.save_screenshot(dir_snapshot + os.path.sep + "screenshot.png")

    main_scroll_height = 0
    main_scroll_top = 0
    page_height = 0
    flag_main_element = False
    try:
        main_scroll_height = driver.execute_script(
            "return document.getElementById('MAIN').scrollHeight"
        )
        main_scroll_top = driver.execute_script(
            "return document.getElementById('MAIN').scrollTop"
        )
        page_height = driver.execute_script("return document.body.scrollHeight")
        flag_main_element = True
    except Exception as e:
        print(e)
        main_scroll_height = driver.execute_script(
            "return document.body.parentNode.scrollHeight"
        )
        main_scroll_top = driver.execute_script(
            "return document.body.parentNode.scrollTop"
        )
        page_height = driver.execute_script("return window.innerHeight")

    print(
        "main_scroll_height is %d main_scroll_top is %d page_height is %d"
        % (main_scroll_height, main_scroll_top, page_height)
    )
    while True:
        driver.execute_script("window.scrollBy(0, %d);" % page_height)
        time.sleep(1)
        if flag_main_element:
            new_main_scroll_height = driver.execute_script(
                "return document.getElementById('MAIN').scrollHeight"
            )
        else:
            new_main_scroll_height = driver.execute_script(
                "return document.body.parentNode.scrollHeight"
            )
        if new_main_scroll_height == main_scroll_height:
            break
        main_scroll_height = new_main_scroll_height
        print("main_scroll_height is %d" % main_scroll_height)

    driver.execute_script("window.scrollTo(0, 0);")
    page_index = 1
    driver.save_screenshot(dir_snapshot + os.path.sep + ("page_%06d.png" % page_index))

    if flag_main_element:
        main_scroll_height = driver.execute_script(
            "return document.getElementById('MAIN').scrollHeight"
        )
        page_height = driver.execute_script("return document.body.scrollHeight")
    else:
        main_scroll_height = driver.execute_script(
            "return document.documentElement.scrollHeight"
        )
        page_height = driver.execute_script("return window.innerHeight")

    window_page_y_offset = driver.execute_script("return window.pageYOffset")
    print(
        "page_index is %d window_page_y_offset is %d main_scroll_height is %d page_height is %d"
        % (page_index, window_page_y_offset, main_scroll_height, page_height)
    )

    while (window_page_y_offset + page_height < main_scroll_height) and (
        page_index < 100
    ):
        driver.execute_script("window.scrollBy(0, %d);" % page_height)
        page_index = page_index + 1
        driver.save_screenshot(
            dir_snapshot + os.path.sep + ("page_%06d.png" % page_index)
        )
        window_page_y_offset = driver.execute_script("return window.pageYOffset")
        print(
            "page_index is %d window_page_y_offset is %d main_scroll_height is %d page_height is %d"
            % (page_index, window_page_y_offset, main_scroll_height, page_height)
        )
    assert page_index < 100
    driver.save_screenshot(dir_snapshot + os.path.sep + "screenshot_last_page.png")
    driver.close()
    return page_index


def find_line_in_file(file_path, line_content, search_line_number_begin=0):
    import fileinput

    line_number = 0
    with fileinput.FileInput(file_path, mode="r") as f:
        for line in f:
            line_number = line_number + 1
            if line_number < search_line_number_begin:
                continue
            if line_content == line:
                return line_number
    return 0
